@@ -2,7 +2,10 @@ package revisions
 
 import (
 	"bytes"
+	"context"
+	"errors"
 	"fmt"
+	"io"
 	"log"
 	"reflect"
 	"slices"
@@ -10,6 +13,7 @@ import (
 	"sync/atomic"
 
 	"github.com/idursun/jjui/internal/ui/common/list"
+	"github.com/idursun/jjui/internal/ui/intents"
 	"github.com/idursun/jjui/internal/ui/operations/ace_jump"
 	"github.com/idursun/jjui/internal/ui/operations/duplicate"
 	"github.com/idursun/jjui/internal/ui/operations/revert"
@@ -71,6 +75,7 @@ type Model struct {
 	textStyle        lipgloss.Style
 	dimmedStyle      lipgloss.Style
 	selectedStyle    lipgloss.Style
+	matchedStyle     lipgloss.Style
 	ensureCursorView bool
 	requestInFlight  bool
 }
@@ -186,6 +191,7 @@ func (m *Model) GetItemRenderer(index int) list.IItemRenderer {
 		textStyle:     m.textStyle,
 		dimmedStyle:   m.dimmedStyle,
 		selectedStyle: m.selectedStyle,
+		matchedStyle:  m.matchedStyle,
 		isChecked:     m.renderer.selections[row.Commit.GetChangeId()],
 		isGutterInLane: func(lineIndex, segmentIndex int) bool {
 			return m.renderer.tracer.IsGutterInLane(index, lineIndex, segmentIndex)
@@ -281,8 +287,8 @@ func (m *Model) Update(msg tea.Msg) tea.Cmd {
 
 func (m *Model) internalUpdate(msg tea.Msg) tea.Cmd {
 	switch msg := msg.(type) {
-	//case rebase.updateHighlightedIdsMsg:
-	//	return m.op.Update(msg)
+	case intents.Intent:
+		return m.handleIntent(msg)
 	case tea.MouseMsg:
 		switch msg.Action {
 		case tea.MouseActionPress:
@@ -304,7 +310,7 @@ func (m *Model) internalUpdate(msg tea.Msg) tea.Cmd {
 		m.op = operations.NewDefault()
 		return m.updateSelection()
 	case common.QuickSearchMsg:
-		m.quickSearch = string(msg)
+		m.quickSearch = strings.ToLower(string(msg))
 		m.SetCursor(m.search(0))
 		m.op = operations.NewDefault()
 		m.renderer.Reset()
@@ -321,18 +327,14 @@ func (m *Model) internalUpdate(msg tea.Msg) tea.Cmd {
 			m.previousOpLogId = currentOperationId
 			return common.RefreshAndKeepSelections
 		}
+	case common.UpdateRevisionsFailedMsg:
+		m.isLoading = false
+		return nil
 	case common.RefreshMsg:
-		if !msg.KeepSelections {
-			m.context.ClearCheckedItems(reflect.TypeFor[appContext.SelectedRevision]())
-		}
-		m.isLoading = true
-		cmd := m.op.Update(msg)
-		if config.Current.Revisions.LogBatching {
-			currentTag := m.tag.Add(1)
-			return tea.Batch(m.loadStreaming(m.context.CurrentRevset, msg.SelectedRevision, currentTag), cmd)
-		} else {
-			return tea.Batch(m.load(m.context.CurrentRevset, msg.SelectedRevision), cmd)
-		}
+		return tea.Batch(m.refresh(intents.Refresh{
+			KeepSelections:   msg.KeepSelections,
+			SelectedRevision: msg.SelectedRevision,
+		}), m.op.Update(msg))
 	case updateRevisionsMsg:
 		m.isLoading = false
 		m.updateGraphRows(msg.rows, msg.selectedRevision)
@@ -352,7 +354,7 @@ func (m *Model) internalUpdate(msg tea.Msg) tea.Cmd {
 				m.revisionToSelect = selected.CommitId
 			}
 		}
-		log.Println("Starting streaming revisions message received with tag:", msg.tag, "revision to select:", msg.selectedRevision)
+		log.Println("Starting streaming revisions with tag:", msg.tag)
 		return m.requestMoreRows(msg.tag)
 	case appendRowsBatchMsg:
 		m.requestInFlight = false
@@ -389,14 +391,12 @@ func (m *Model) internalUpdate(msg tea.Msg) tea.Cmd {
 		}
 
 		cmds := []tea.Cmd{m.highlightChanges, m.updateSelection()}
-		if !m.hasMore {
+		if len(m.offScreenRows) > 0 {
 			cmds = append(cmds, func() tea.Msg {
 				return common.UpdateRevisionsSuccessMsg{}
 			})
 		}
 		return tea.Batch(cmds...)
-	case common.StartSquashOperationMsg:
-		return m.startSquash(jj.NewSelectedRevisions(msg.Revision), msg.Files)
 	}
 
 	if len(m.rows) == 0 {
@@ -416,58 +416,15 @@ func (m *Model) internalUpdate(msg tea.Msg) tea.Cmd {
 	case tea.KeyMsg:
 		switch {
 		case key.Matches(msg, m.keymap.Up, pageUpKey):
-			amount := 1
-			if key.Matches(msg, pageUpKey) && m.renderer.LastRowIndex > m.renderer.FirstRowIndex {
-				amount = m.renderer.LastRowIndex - m.renderer.FirstRowIndex - 1
-			}
-			if m.cursor == 0 && key.Matches(msg, pageUpKey) {
-				return func() tea.Msg {
-					return common.CommandCompletedMsg{
-						Output: fmt.Sprintf("Already at the top of revset `%s`", m.context.CurrentRevset),
-						Err:    nil,
-					}
-				}
-			}
-			m.SetCursor(max(m.cursor-amount, 0))
-			return m.updateSelection()
+			return m.handleIntent(intents.Navigate{Delta: -1, Page: key.Matches(msg, pageUpKey)})
 		case key.Matches(msg, m.keymap.Down, pageDownKey):
-			amount := 1
-			if key.Matches(msg, pageDownKey) && m.renderer.LastRowIndex > m.renderer.FirstRowIndex {
-				amount = m.renderer.LastRowIndex - m.renderer.FirstRowIndex - 1
-			}
-			if len(m.rows) > 0 && m.cursor == len(m.rows)-1 && !m.hasMore && key.Matches(msg, pageDownKey) {
-				return func() tea.Msg {
-					return common.CommandCompletedMsg{
-						Output: fmt.Sprintf("Already at the bottom of revset `%s`", m.context.CurrentRevset),
-						Err:    nil,
-					}
-				}
-			}
-			if m.cursor < len(m.rows)-amount {
-				m.SetCursor(m.cursor + amount)
-			} else if m.hasMore {
-				return m.requestMoreRows(m.tag.Load())
-			} else if len(m.rows) > 0 {
-				m.SetCursor(len(m.rows) - 1)
-			}
-			m.ensureCursorView = true
-			return m.updateSelection()
+			return m.handleIntent(intents.Navigate{Delta: 1, Page: key.Matches(msg, pageDownKey)})
 		case key.Matches(msg, m.keymap.JumpToParent):
-			m.jumpToParent(m.SelectedRevisions())
-			return m.updateSelection()
+			return m.handleIntent(intents.Navigate{Target: intents.TargetParent})
 		case key.Matches(msg, m.keymap.JumpToChildren):
-			immediate, _ := m.context.RunCommandImmediate(jj.GetFirstChild(m.SelectedRevision()))
-			index := m.selectRevision(string(immediate))
-			if index != -1 {
-				m.SetCursor(index)
-			}
-			return m.updateSelection()
+			return m.handleIntent(intents.Navigate{Target: intents.TargetChild})
 		case key.Matches(msg, m.keymap.JumpToWorkingCopy):
-			workingCopyIndex := m.selectRevision("@")
-			if workingCopyIndex != -1 {
-				m.SetCursor(workingCopyIndex)
-			}
-			return m.updateSelection()
+			return m.handleIntent(intents.Navigate{Target: intents.TargetWorkingCopy})
 		case key.Matches(msg, m.keymap.AceJump):
 			op := ace_jump.NewOperation(m, func(index int) parser.Row {
 				return m.rows[index]
@@ -480,86 +437,67 @@ func (m *Model) internalUpdate(msg tea.Msg) tea.Cmd {
 			}
 
 			switch {
+			case m.quickSearch != "" && (msg.Type == tea.KeyEsc || msg.Type == tea.KeyEnter):
+				m.quickSearch = ""
+				m.renderer.Reset()
+				return nil
 			case key.Matches(msg, m.keymap.ToggleSelect):
 				commit := m.rows[m.cursor].Commit
 				changeId := commit.GetChangeId()
 				item := appContext.SelectedRevision{ChangeId: changeId, CommitId: commit.CommitId}
 				m.context.ToggleCheckedItem(item)
-				immediate, _ := m.context.RunCommandImmediate(jj.GetParent(jj.NewSelectedRevisions(commit)))
-				parentIndex := m.selectRevision(string(immediate))
-				if parentIndex != -1 {
-					m.SetCursor(parentIndex)
-				}
+				m.jumpToParent(jj.NewSelectedRevisions(commit))
 			case key.Matches(msg, m.keymap.Cancel):
+				m.context.ClearCheckedItems(reflect.TypeFor[appContext.SelectedRevision]())
 				m.op = operations.NewDefault()
+				m.renderer.Reset()
+				return nil
 			case key.Matches(msg, m.keymap.QuickSearchCycle):
 				m.SetCursor(m.search(m.cursor + 1))
 				m.renderer.Reset()
 				return nil
 			case key.Matches(msg, m.keymap.Details.Mode):
-				model := details.NewOperation(m.context, m.SelectedRevision())
-				model.Parent = m.ViewNode
-				m.op = model
-				return m.op.Init()
+				return m.handleIntent(intents.OpenDetails{})
 			case key.Matches(msg, m.keymap.InlineDescribe.Mode):
-				model := describe.NewOperation(m.context, m.SelectedRevision().GetChangeId())
-				model.Parent = m.ViewNode
-				m.op = model
-				return m.op.Init()
+				return m.handleIntent(intents.StartInlineDescribe{})
 			case key.Matches(msg, m.keymap.New):
-				return m.context.RunCommand(jj.New(m.SelectedRevisions()), common.RefreshAndSelect("@"))
+				return m.handleIntent(intents.StartNew{})
 			case key.Matches(msg, m.keymap.Commit):
-				return m.context.RunInteractiveCommand(jj.CommitWorkingCopy(), common.Refresh)
+				return m.handleIntent(intents.CommitWorkingCopy{})
 			case key.Matches(msg, m.keymap.Edit, m.keymap.ForceEdit):
 				ignoreImmutable := key.Matches(msg, m.keymap.ForceEdit)
-				return m.context.RunCommand(jj.Edit(m.SelectedRevision().GetChangeId(), ignoreImmutable), common.Refresh)
+				return m.handleIntent(intents.StartEdit{IgnoreImmutable: ignoreImmutable})
 			case key.Matches(msg, m.keymap.Diffedit):
-				changeId := m.SelectedRevision().GetChangeId()
-				return m.context.RunInteractiveCommand(jj.DiffEdit(changeId), common.Refresh)
+				return m.handleIntent(intents.StartDiffEdit{})
 			case key.Matches(msg, m.keymap.Absorb):
-				changeId := m.SelectedRevision().GetChangeId()
-				return m.context.RunCommand(jj.Absorb(changeId), common.Refresh)
+				return m.handleIntent(intents.StartAbsorb{})
 			case key.Matches(msg, m.keymap.Abandon):
-				selections := m.SelectedRevisions()
-				m.op = abandon.NewOperation(m.context, selections)
-				return m.op.Init()
+				return m.handleIntent(intents.StartAbandon{})
 			case key.Matches(msg, m.keymap.Bookmark.Set):
 				m.op = bookmark.NewSetBookmarkOperation(m.context, m.SelectedRevision().GetChangeId())
 				return m.op.Init()
 			case key.Matches(msg, m.keymap.Split, m.keymap.SplitParallel):
-				isParallel := key.Matches(msg, m.keymap.SplitParallel)
-				currentRevision := m.SelectedRevision().GetChangeId()
-				return m.context.RunInteractiveCommand(jj.Split(currentRevision, []string{}, isParallel), common.Refresh)
+				return m.handleIntent(intents.StartSplit{
+					IsParallel: key.Matches(msg, m.keymap.SplitParallel),
+				})
 			case key.Matches(msg, m.keymap.Describe):
-				selections := m.SelectedRevisions()
-				return m.context.RunInteractiveCommand(jj.Describe(selections), common.Refresh)
+				return m.handleIntent(intents.StartDescribe{})
 			case key.Matches(msg, m.keymap.Evolog.Mode):
-				model := evolog.NewOperation(m.context, m.SelectedRevision())
-				model.Parent = m.ViewNode
-				m.op = model
-				return m.op.Init()
+				return m.handleIntent(intents.StartEvolog{})
 			case key.Matches(msg, m.keymap.Diff):
-				return func() tea.Msg {
-					changeId := m.SelectedRevision().GetChangeId()
-					output, _ := m.context.RunCommandImmediate(jj.Diff(changeId, ""))
-					return common.ShowDiffMsg(output)
-				}
+				return m.handleIntent(intents.ShowDiff{})
 			case key.Matches(msg, m.keymap.Refresh):
-				return common.Refresh
+				return m.handleIntent(intents.Refresh{})
 			case key.Matches(msg, m.keymap.Squash.Mode):
-				return m.startSquash(m.SelectedRevisions(), nil)
+				return m.handleIntent(intents.StartSquash{})
 			case key.Matches(msg, m.keymap.Revert.Mode):
-				m.op = revert.NewOperation(m.context, m.SelectedRevisions(), revert.TargetDestination)
-				return m.op.Init()
+				return m.handleIntent(intents.StartRevert{})
 			case key.Matches(msg, m.keymap.Rebase.Mode):
-				m.op = rebase.NewOperation(m.context, m.SelectedRevisions(), rebase.SourceRevision, rebase.TargetDestination)
-				return m.op.Init()
+				return m.handleIntent(intents.StartRebase{})
 			case key.Matches(msg, m.keymap.Duplicate.Mode):
-				m.op = duplicate.NewOperation(m.context, m.SelectedRevisions(), duplicate.TargetDestination)
-				return m.op.Init()
+				return m.handleIntent(intents.StartDuplicate{})
 			case key.Matches(msg, m.keymap.SetParents):
-				m.op = set_parents.NewModel(m.context, m.SelectedRevision())
-				return m.op.Init()
+				return m.handleIntent(intents.SetParents{})
 			}
 		}
 	}
@@ -567,19 +505,379 @@ func (m *Model) internalUpdate(msg tea.Msg) tea.Cmd {
 	return cmd
 }
 
-func (m *Model) startSquash(selectedRevisions jj.SelectedRevisions, files []string) tea.Cmd {
-	parent, _ := m.context.RunCommandImmediate(jj.GetParent(selectedRevisions))
+func (m *Model) handleIntent(intent intents.Intent) tea.Cmd {
+	switch intent := intent.(type) {
+	case intents.OpenDetails:
+		return m.openDetails(intent)
+	case intents.StartSquash:
+		return m.startSquash(intent)
+	case intents.StartInlineDescribe:
+		return m.startInlineDescribe(intent)
+	case intents.StartAbsorb:
+		return m.startAbsorb(intent)
+	case intents.StartAbandon:
+		return m.startAbandon(intent)
+	case intents.StartNew:
+		return m.startNew(intent)
+	case intents.CommitWorkingCopy:
+		return m.commitWorkingCopy()
+	case intents.StartEdit:
+		return m.startEdit(intent)
+	case intents.StartDiffEdit:
+		return m.startDiffEdit(intent)
+	case intents.StartRevert:
+		return m.startRevert(intent)
+	case intents.StartDuplicate:
+		return m.startDuplicate(intent)
+	case intents.SetParents:
+		return m.startSetParents(intent)
+	case intents.Navigate:
+		return m.navigate(intent)
+	case intents.StartDescribe:
+		return m.startDescribe(intent)
+	case intents.StartEvolog:
+		return m.startEvolog(intent)
+	case intents.ShowDiff:
+		return m.showDiff(intent)
+	case intents.StartSplit:
+		return m.startSplit(intent)
+	case intents.StartRebase:
+		return m.startRebase(intent)
+	case intents.Refresh:
+		return m.refresh(intent)
+	}
+	return nil
+}
+
+func (m *Model) refresh(intent intents.Refresh) tea.Cmd {
+	if !intent.KeepSelections {
+		m.context.ClearCheckedItems(reflect.TypeFor[appContext.SelectedRevision]())
+	}
+	m.isLoading = true
+	if config.Current.Revisions.LogBatching {
+		currentTag := m.tag.Add(1)
+		return m.loadStreaming(m.context.CurrentRevset, intent.SelectedRevision, currentTag)
+	}
+	return m.load(m.context.CurrentRevset, intent.SelectedRevision)
+}
+
+func (m *Model) openDetails(_ intents.OpenDetails) tea.Cmd {
+	if m.SelectedRevision() == nil {
+		return nil
+	}
+	model := details.NewOperation(m.context, m.SelectedRevision())
+	model.Parent = m.ViewNode
+	m.op = model
+	return m.op.Init()
+}
+
+func (m *Model) startSquash(intent intents.StartSquash) tea.Cmd {
+	selected := intent.Selected
+	if len(selected.Revisions) == 0 {
+		selected = m.SelectedRevisions()
+	}
+	if len(selected.Revisions) == 0 {
+		return nil
+	}
+
+	parent, _ := m.context.RunCommandImmediate(jj.GetParent(selected))
 	parentIdx := m.selectRevision(string(parent))
 	if parentIdx != -1 {
 		m.SetCursor(parentIdx)
 	} else if m.cursor < len(m.rows)-1 {
 		m.SetCursor(m.cursor + 1)
 	}
-	m.op = squash.NewOperation(m.context, selectedRevisions, squash.WithFiles(files))
+	m.op = squash.NewOperation(m.context, selected, squash.WithFiles(intent.Files))
 	return m.op.Init()
 }
 
+func (m *Model) startRebase(intent intents.StartRebase) tea.Cmd {
+	selected := intent.Selected
+	if len(selected.Revisions) == 0 {
+		selected = m.SelectedRevisions()
+	}
+	if len(selected.Revisions) == 0 {
+		return nil
+	}
+
+	source := intent.Source
+	if source == 0 {
+		source = rebase.SourceRevision
+	}
+	target := intent.Target
+	if target == 0 {
+		target = rebase.TargetDestination
+	}
+	m.op = rebase.NewOperation(m.context, selected, source, target)
+	return m.op.Init()
+}
+
+func (m *Model) startRevert(intent intents.StartRevert) tea.Cmd {
+	selected := intent.Selected
+	if len(selected.Revisions) == 0 {
+		selected = m.SelectedRevisions()
+	}
+	if len(selected.Revisions) == 0 {
+		return nil
+	}
+
+	target := intent.Target
+	if target == 0 {
+		target = revert.TargetDestination
+	}
+	m.op = revert.NewOperation(m.context, selected, target)
+	return m.op.Init()
+}
+
+func (m *Model) startDuplicate(intent intents.StartDuplicate) tea.Cmd {
+	selected := intent.Selected
+	if len(selected.Revisions) == 0 {
+		selected = m.SelectedRevisions()
+	}
+	if len(selected.Revisions) == 0 {
+		return nil
+	}
+
+	m.op = duplicate.NewOperation(m.context, selected, duplicate.TargetDestination)
+	return m.op.Init()
+}
+
+func (m *Model) startSetParents(intent intents.SetParents) tea.Cmd {
+	commit := intent.Selected
+	if commit == nil {
+		commit = m.SelectedRevision()
+	}
+	if commit == nil {
+		return nil
+	}
+
+	m.op = set_parents.NewModel(m.context, commit)
+	return m.op.Init()
+}
+
+func (m *Model) startNew(intent intents.StartNew) tea.Cmd {
+	selected := intent.Selected
+	if len(selected.Revisions) == 0 {
+		selected = m.SelectedRevisions()
+	}
+	return m.context.RunCommand(jj.New(selected), common.RefreshAndSelect("@"))
+}
+
+func (m *Model) commitWorkingCopy() tea.Cmd {
+	return m.context.RunInteractiveCommand(jj.CommitWorkingCopy(), common.Refresh)
+}
+
+func (m *Model) startEdit(intent intents.StartEdit) tea.Cmd {
+	commit := intent.Selected
+	if commit == nil {
+		commit = m.SelectedRevision()
+	}
+	if commit == nil {
+		return nil
+	}
+	return m.context.RunCommand(jj.Edit(commit.GetChangeId(), intent.IgnoreImmutable), common.Refresh)
+}
+
+func (m *Model) startDiffEdit(intent intents.StartDiffEdit) tea.Cmd {
+	commit := intent.Selected
+	if commit == nil {
+		commit = m.SelectedRevision()
+	}
+	if commit == nil {
+		return nil
+	}
+	return m.context.RunInteractiveCommand(jj.DiffEdit(commit.GetChangeId()), common.Refresh)
+}
+
+func (m *Model) startAbsorb(intent intents.StartAbsorb) tea.Cmd {
+	commit := intent.Selected
+	if commit == nil {
+		commit = m.SelectedRevision()
+	}
+	if commit == nil {
+		return nil
+	}
+	return m.context.RunCommand(jj.Absorb(commit.GetChangeId()), common.Refresh)
+}
+
+func (m *Model) startAbandon(intent intents.StartAbandon) tea.Cmd {
+	selected := intent.Selected
+	if len(selected.Revisions) == 0 {
+		selected = m.SelectedRevisions()
+	}
+	if len(selected.Revisions) == 0 {
+		return nil
+	}
+	m.op = abandon.NewOperation(m.context, selected)
+	return m.op.Init()
+}
+
+func (m *Model) navigate(intent intents.Navigate) tea.Cmd {
+	if len(m.rows) == 0 {
+		return nil
+	}
+
+	ensureView := true
+	if intent.EnsureView != nil {
+		ensureView = *intent.EnsureView
+	}
+	allowStream := true
+	if intent.AllowStream != nil {
+		allowStream = *intent.AllowStream
+	}
+
+	if intent.ChangeID != "" || intent.FallbackID != "" {
+		idx := m.selectRevision(intent.ChangeID)
+		if idx == -1 && intent.FallbackID != "" {
+			idx = m.selectRevision(intent.FallbackID)
+		}
+		if idx == -1 {
+			return nil
+		}
+		m.ensureCursorView = ensureView
+		m.SetCursor(idx)
+		return m.updateSelection()
+	}
+
+	switch intent.Target {
+	case intents.TargetParent:
+		m.jumpToParent(m.SelectedRevisions())
+		m.ensureCursorView = ensureView
+		return m.updateSelection()
+	case intents.TargetWorkingCopy:
+		if idx := m.selectRevision("@"); idx != -1 {
+			m.SetCursor(idx)
+		}
+		m.ensureCursorView = ensureView
+		return m.updateSelection()
+	case intents.TargetChild:
+		immediate, _ := m.context.RunCommandImmediate(jj.GetFirstChild(m.SelectedRevision()))
+		if idx := m.selectRevision(string(immediate)); idx != -1 {
+			m.SetCursor(idx)
+		}
+		m.ensureCursorView = ensureView
+		return m.updateSelection()
+	}
+
+	step := intent.Delta
+	if step == 0 {
+		step = 1
+	}
+	if intent.Page {
+		span := m.renderer.LastRowIndex - m.renderer.FirstRowIndex - 1
+		if span < 1 {
+			span = 1
+		}
+		if step < 0 {
+			step = -span
+		} else {
+			step = span
+		}
+	}
+
+	if step > 0 {
+		if len(m.rows) > 0 && m.cursor == len(m.rows)-1 && !m.hasMore && step > 1 {
+			return func() tea.Msg {
+				return common.CommandCompletedMsg{
+					Output: fmt.Sprintf("Already at the bottom of revset `%s`", m.context.CurrentRevset),
+					Err:    nil,
+				}
+			}
+		}
+		if m.cursor+step < len(m.rows) {
+			m.SetCursor(m.cursor + step)
+		} else if allowStream && m.hasMore {
+			return m.requestMoreRows(m.tag.Load())
+		} else if len(m.rows) > 0 {
+			m.SetCursor(len(m.rows) - 1)
+		}
+	} else if step < 0 {
+		amount := -step
+		if m.cursor == 0 && amount > 1 {
+			return func() tea.Msg {
+				return common.CommandCompletedMsg{
+					Output: fmt.Sprintf("Already at the top of revset `%s`", m.context.CurrentRevset),
+					Err:    nil,
+				}
+			}
+		}
+		m.SetCursor(max(m.cursor-amount, 0))
+	}
+
+	m.ensureCursorView = ensureView
+	return m.updateSelection()
+}
+
+func (m *Model) startDescribe(intent intents.StartDescribe) tea.Cmd {
+	selected := intent.Selected
+	if len(selected.Revisions) == 0 {
+		selected = m.SelectedRevisions()
+	}
+	if len(selected.Revisions) == 0 {
+		return nil
+	}
+	return m.context.RunInteractiveCommand(jj.Describe(selected), common.Refresh)
+}
+
+func (m *Model) startEvolog(intent intents.StartEvolog) tea.Cmd {
+	commit := intent.Selected
+	if commit == nil {
+		commit = m.SelectedRevision()
+	}
+	if commit == nil {
+		return nil
+	}
+	model := evolog.NewOperation(m.context, commit)
+	model.Parent = m.ViewNode
+	m.op = model
+	return m.op.Init()
+}
+
+func (m *Model) startInlineDescribe(intent intents.StartInlineDescribe) tea.Cmd {
+	commit := intent.Selected
+	if commit == nil {
+		commit = m.SelectedRevision()
+	}
+	if commit == nil {
+		return nil
+	}
+	model := describe.NewOperation(m.context, commit)
+	model.Parent = m.ViewNode
+	m.op = model
+	return m.op.Init()
+}
+
+func (m *Model) showDiff(intent intents.ShowDiff) tea.Cmd {
+	commit := intent.Selected
+	if commit == nil {
+		commit = m.SelectedRevision()
+	}
+	if commit == nil {
+		return nil
+	}
+	changeId := commit.GetChangeId()
+	return func() tea.Msg {
+		output, _ := m.context.RunCommandImmediate(jj.Diff(changeId, ""))
+		return common.ShowDiffMsg(output)
+	}
+}
+
+func (m *Model) startSplit(intent intents.StartSplit) tea.Cmd {
+	commit := intent.Selected
+	if commit == nil {
+		commit = m.SelectedRevision()
+	}
+	if commit == nil {
+		return nil
+	}
+	return m.context.RunInteractiveCommand(jj.Split(commit.GetChangeId(), intent.Files, intent.IsParallel), common.Refresh)
+}
+
 func (m *Model) updateSelection() tea.Cmd {
+	// Don't override file-level selections (from Details panel)
+	if _, isFile := m.context.SelectedItem.(appContext.SelectedFile); isFile && !m.InNormalMode() {
+		return nil
+	}
 	if selectedRevision := m.SelectedRevision(); selectedRevision != nil {
 		return m.context.SetSelectedItem(appContext.SelectedRevision{
 			ChangeId: selectedRevision.GetChangeId(),
@@ -682,14 +980,23 @@ func (m *Model) loadStreaming(revset string, selectedRevision string, tag uint64
 		return nil
 	}
 
-	streamer, err := graph.NewGraphStreamer(m.context, revset, m.context.JJConfig.Templates.Log)
+	var cmds []tea.Cmd
+	streamer, err := graph.NewGraphStreamer(context.Background(), m.context, revset, m.context.JJConfig.Templates.Log)
 	if err != nil {
-		return func() tea.Msg {
+		var errMsg string
+		if err == io.EOF {
+			errMsg = fmt.Sprintf("No revisions found for revset `%s`", revset)
+			err = errors.New(errMsg)
+		} else {
+			errMsg = fmt.Sprintf("%v", err)
+		}
+
+		cmds = append(cmds, func() tea.Msg {
 			return common.UpdateRevisionsFailedMsg{
 				Err:    err,
-				Output: fmt.Sprintf("%v", err),
+				Output: errMsg,
 			}
-		}
+		})
 	}
 
 	if m.streamer != nil {
@@ -698,8 +1005,8 @@ func (m *Model) loadStreaming(revset string, selectedRevision string, tag uint64
 	}
 	m.streamer = streamer
 	m.hasMore = true
-
-	return m.Update(startRowsStreamingMsg{selectedRevision: selectedRevision, tag: tag})
+	cmds = append(cmds, m.Update(startRowsStreamingMsg{selectedRevision: selectedRevision, tag: tag}))
+	return tea.Batch(cmds...)
 }
 
 func (m *Model) requestMoreRows(tag uint64) tea.Cmd {
@@ -709,8 +1016,7 @@ func (m *Model) requestMoreRows(tag uint64) tea.Cmd {
 
 	m.requestInFlight = true
 	batch := m.streamer.RequestMore()
-	m.Update(appendRowsBatchMsg{batch.Rows, batch.HasMore, tag})
-	return nil
+	return m.Update(appendRowsBatchMsg{batch.Rows, batch.HasMore, tag})
 }
 
 func (m *Model) selectRevision(revision string) int {
@@ -738,7 +1044,7 @@ func (m *Model) search(startIndex int) int {
 		row := &m.rows[c]
 		for _, line := range row.Lines {
 			for _, segment := range line.Segments {
-				if segment.Text != "" && strings.Contains(segment.Text, m.quickSearch) {
+				if segment.Text != "" && strings.Contains(strings.ToLower(segment.Text), m.quickSearch) {
 					return c
 				}
 			}
@@ -773,6 +1079,7 @@ func New(c *appContext.MainContext) *Model {
 		textStyle:     common.DefaultPalette.Get("revisions text"),
 		dimmedStyle:   common.DefaultPalette.Get("revisions dimmed"),
 		selectedStyle: common.DefaultPalette.Get("revisions selected"),
+		matchedStyle:  common.DefaultPalette.Get("revisions matched"),
 	}
 	m.renderer = newRevisionListRenderer(&m, m.ViewNode)
 	return &m
